@@ -317,6 +317,7 @@ ngx_http_push_stream_convert_char_to_msg_on_shared(ngx_http_push_stream_main_con
         ngx_http_push_stream_free_message_memory(shpool, msg);
         return NULL;
     }
+    ngx_memzero(msg->formatted_messages, sizeof(ngx_str_t) * msg->qtd_templates);
 
     for (q = ngx_queue_head(&mcf->msg_templates); q != ngx_queue_sentinel(&mcf->msg_templates); q = ngx_queue_next(q)) {
         ngx_http_push_stream_template_t *cur = ngx_queue_data(q, ngx_http_push_stream_template_t, queue);
@@ -585,18 +586,19 @@ ngx_http_push_stream_send_response_message(ngx_http_request_t *r, ngx_http_push_
                 }
             }
 
-            if (rc == NGX_OK) {
-                rc = ngx_http_push_stream_send_response_text(r, str->data, str->len, 0);
+            if ((rc == NGX_OK) && use_jsonp && send_separator) {
+                rc = ngx_http_push_stream_send_response_text(r, NGX_HTTP_PUSH_STREAM_CALLBACK_MID_CHUNK.data, NGX_HTTP_PUSH_STREAM_CALLBACK_MID_CHUNK.len, 0);
             }
 
-            if ((rc == NGX_OK) && use_jsonp) {
-                if (send_separator) {
-                    rc = ngx_http_push_stream_send_response_text(r, NGX_HTTP_PUSH_STREAM_CALLBACK_MID_CHUNK.data, NGX_HTTP_PUSH_STREAM_CALLBACK_MID_CHUNK.len, 0);
+            if (rc == NGX_OK) {
+                rc = ngx_http_push_stream_send_response_text(r, str->data, str->len, 0);
+                if (rc == NGX_OK) {
+                    ctx->message_sent = 1;
                 }
+            }
 
-                if (send_callback) {
-                    rc = ngx_http_push_stream_send_response_text(r, NGX_HTTP_PUSH_STREAM_CALLBACK_END_CHUNK.data, NGX_HTTP_PUSH_STREAM_CALLBACK_END_CHUNK.len, 0);
-                }
+            if ((rc == NGX_OK) && use_jsonp && send_callback) {
+                rc = ngx_http_push_stream_send_response_text(r, NGX_HTTP_PUSH_STREAM_CALLBACK_END_CHUNK.data, NGX_HTTP_PUSH_STREAM_CALLBACK_END_CHUNK.len, 0);
             }
 
             if (rc == NGX_OK) {
@@ -1370,8 +1372,9 @@ ngx_http_push_stream_format_message(ngx_http_push_stream_channel_t *channel, ngx
     ngx_queue_t               *q;
     u_char                     id[NGX_INT_T_LEN + 1];
     u_char                     tag[NGX_INT_T_LEN + 1];
+    u_char                     size[NGX_INT_T_LEN + 1];
     u_char                     time[NGX_HTTP_PUSH_STREAM_TIME_FMT_LEN + 1];
-    size_t                     id_len, tag_len, time_len;
+    size_t                     id_len, tag_len, time_len, size_len;
 
     ngx_str_t *channel_id = (channel != NULL) ? &channel->id : &NGX_HTTP_PUSH_STREAM_EMPTY;
     ngx_str_t *event_id = (message->event_id != NULL) ? message->event_id : &NGX_HTTP_PUSH_STREAM_EMPTY;
@@ -1386,6 +1389,9 @@ ngx_http_push_stream_format_message(ngx_http_push_stream_channel_t *channel, ngx
     ngx_sprintf(tag, "%d%Z", message->tag);
     tag_len = ngx_strlen(tag);
 
+    ngx_sprintf(size, "%d%Z", text->len);
+    size_len = ngx_strlen(size);
+
     len += template->qtd_channel * channel_id->len;
     len += template->qtd_event_id * event_id->len;
     len += template->qtd_event_type * event_type->len;
@@ -1393,6 +1399,7 @@ ngx_http_push_stream_format_message(ngx_http_push_stream_channel_t *channel, ngx
     len += template->qtd_time * time_len;
     len += template->qtd_tag * tag_len;
     len += template->qtd_text * text->len;
+    len += template->qtd_size * size_len;
     len += template->literal_len;
 
     txt = ngx_http_push_stream_create_str(temp_pool, len);
@@ -1426,6 +1433,9 @@ ngx_http_push_stream_format_message(ngx_http_push_stream_channel_t *channel, ngx
             case PUSH_STREAM_TEMPLATE_PART_TYPE_TEXT:
                 last = ngx_cpymem(last, text->data, text->len);
                 break;
+            case PUSH_STREAM_TEMPLATE_PART_TYPE_SIZE:
+                last = ngx_cpymem(last, size, size_len);
+                break;
             case PUSH_STREAM_TEMPLATE_PART_TYPE_TIME:
                 last = ngx_cpymem(last, time, time_len);
                 break;
@@ -1467,6 +1477,7 @@ ngx_http_push_stream_add_request_context(ngx_http_request_t *r)
     ctx->ping_timer = NULL;
     ctx->subscriber = NULL;
     ctx->longpolling = 0;
+    ctx->message_sent = 0;
     ctx->padding = NULL;
     ctx->callback = NULL;
     ctx->requested_channels = NULL;
@@ -1805,9 +1816,9 @@ ngx_http_push_stream_add_polling_headers(ngx_http_request_t *r, time_t last_modi
     }
 
     if (tag >= 0) {
-        ngx_str_t *etag = ngx_http_push_stream_create_str(temp_pool, NGX_INT_T_LEN);
+        ngx_str_t *etag = ngx_http_push_stream_create_str(temp_pool, NGX_INT_T_LEN + 3);
         if (etag != NULL) {
-            ngx_sprintf(etag->data, "%ui", tag);
+            ngx_sprintf(etag->data, "W/%ui%Z", tag);
             etag->len = ngx_strlen(etag->data);
             r->headers_out.etag = ngx_http_push_stream_add_response_header(r, &NGX_HTTP_PUSH_STREAM_HEADER_ETAG, etag);
         }
@@ -1836,6 +1847,11 @@ ngx_http_push_stream_get_last_received_message_values(ngx_http_request_t *r, tim
         etag = ngx_http_push_stream_get_header(r, &NGX_HTTP_PUSH_STREAM_HEADER_IF_NONE_MATCH);
     }
 
+    if ((etag != NULL) && (etag->len > 2) && (etag->data[0] == 'W') && (etag->data[1] == '/')) {
+        etag->len -= 2;
+        etag->data = etag->data + 2;
+    }
+
     if (cf->last_event_id != NULL) {
         ngx_http_push_stream_complex_value(r, cf->last_event_id, &vv_event_id);
         if (vv_event_id.len) {
@@ -2303,3 +2319,30 @@ ngx_http_push_stream_create_shmtx(ngx_shmtx_t *mtx, ngx_shmtx_sh_t *addr, u_char
 
     return NGX_OK;
 }
+
+
+ngx_flag_t
+ngx_http_push_stream_is_utf8(u_char *p, size_t n)
+{
+    u_char  c, *last;
+    size_t  len;
+
+    last = p + n;
+
+    for (len = 0; p < last; len++) {
+
+        c = *p;
+
+        if (c < 0x80) {
+            p++;
+            continue;
+        }
+
+        if (ngx_utf8_decode(&p, n) > 0x10ffff) {
+            /* invalid UTF-8 */
+            return 0;
+        }
+    }
+
+    return 1;
+}
